/******************************************************************************
 * file: SingleLinkedList.java
 * 
 * description: provides a class to create single linked lists and methods
 * to perform typical operations on these type of data set
 *****************************************************************************/

import java.util.Iterator;

/* implements the interface Iterator */
class SingleLinkedListIterator<E> implements Iterator<E> {

	public SingleLinkedListIterator(Node_SLL<E> node) {
		this.node = node;
	}

	public boolean hasNext() {
		return (this.node != null);
	}

	public E next() {
		E e = node.getContent();
		this.node = this.node.getNextNode();
		return e;
	}
	
	public void remove() {
		// this method has to be implemented!
	}

	private Node_SLL<E> node;
}

/* a node of a single linked list has a content and a successor */
class Node_SLL<E> {

	/* constructor */
	public Node_SLL(E o, Node_SLL<E> n) {
		this.content = o;
		this.nextNode = n;
	}

	/* returns the content of this node */
	public E getContent() {
		return this.content;
	}

	/* sets the content of this node to o */
	public void setContent(E o) {
		this.content = o;
	}

	/* returns the successor of this node */
	public Node_SLL<E> getNextNode() {
		return this.nextNode;
	}

	/* sets the successor of this node to n */
	public void setNextNode(Node_SLL<E> n) {
		this.nextNode = n;
	}

	/* returns the node's content as a string */
	public String toString() {
		return this.content.toString();
	}

	/* variables */
	private E content;
	private Node_SLL<E> nextNode;

}

/* a single linked list */
public class SingleLinkedList<E> {

	/* constructor */
	public SingleLinkedList() {
		this.head = null;
	}

	/* returns true if this list is empty, otherwise false */
	public boolean isEmpty() {
		return (this.head == null);
	}

	/* returns the reference of the node, that contains E o */
	/* if this list contains E o, otherwise null */
	public Node_SLL<E> search1(E o) {
		Node_SLL<E> current = this.head;
		while (current != null) {
			if (current.getContent().equals(o)) {
				return current;
			}
			current = current.getNextNode();
		}
		return null;
	}

	/* returns the reference of the E o */
	/* if this list contains E o, otherwise null */

	public boolean contains(E o, SingleLinkedList<E> l) {
		Iterator<E> i = l.iterator();
		while (i.hasNext()) {
			E content = i.next();
			if (content.equals(o)) {
				return true;
			}
		}
		return false;
	}

	public Node_SLL<E> searchNode(E o) {
		Node_SLL<E> n = this.head;
		while (n != null && !n.getContent().equals(o)) {
			n = n.getNextNode();
		}
		return n;
	}

	public void insertAfterNode(E o, Node_SLL<E> n) {
		n.setNextNode(new Node_SLL<E>(o, n.getNextNode()));
		System.out.println("insertingAfter " + o);
	}

	/* inserts a Node that contains E o at the head */
	public void insertFirst(E o) {
		this.head = new Node_SLL<E>(o, this.head);
		System.out.println("insertingFirst " + o);
	}

	/* inserts a Node that contains E o at the tail */
	public void insertLast(E o) {
		if (this.isEmpty()) { /* list is empty */
			this.insertFirst(o);
		} else {
			Node_SLL<E> tmp = this.head;
			while (tmp.getNextNode() != null) {
				tmp = tmp.getNextNode();
			}
			tmp.setNextNode(new Node_SLL<E>(o, null));
			System.out.println("insertingLast " + o);
		}
	}

	/* removes the first Node */
	void removeFirstNode() {
		if (!this.isEmpty()) {
			System.out.println("removing first node");
			this.head = this.head.getNextNode();
		}
	}

	/* removes the next Node after Node */
	private void removeNextNode(Node_SLL<E> node) {
		if (node.getNextNode() != null) {
			System.out.println("removing "
			    + node.getNextNode().getContent().toString());
			node.setNextNode(node.getNextNode().getNextNode());
		}
	}

	/* removes the first Node that contains E o */
	public void remove(E o) {
		Node_SLL<E> current = this.head;
		Node_SLL<E> predecessor = null;
		while (current != null) {
			if (current.getContent().equals(o)) {
				if (predecessor != null) {
					predecessor.setNextNode(current.getNextNode());
				} else {
					this.head = current.getNextNode();
				}
				System.out.println("removing " + o);
				break;
			}
			predecessor = current;
			current = current.getNextNode();
		}
	}

	private void reverseRecursive(Node_SLL<E> node, Node_SLL<E> pred) {
		Node_SLL<E> next = node.getNextNode();
		if (next == null) {
			this.head = node;
		} else {
			reverseRecursive(next, node);
		}
		node.setNextNode(pred);
	}

	/* reverses the list */
	public void reverseList() {
		if (!this.isEmpty()) {
			reverseRecursive(this.head, null);
		}
		System.out.println("reversing list");
	}

	/* returns the list's content as a string object */
	public String toString() {
		String str = "[";
		Node_SLL<E> tmp = this.head;
		while (tmp != null) {
			str += tmp.getContent().toString();
			tmp = tmp.getNextNode();
			if (tmp != null) {
				str += ", ";
			}
		}
		return str + "]";
	}

	public Iterator<E> iterator() {
		return new SingleLinkedListIterator<E>(this.head);
	}

	/* variables */
	private Node_SLL<E> head;
}
